/*
  INF3105 - Structures de donnes et algorithmes
  UQAM / Dpartement d'informatique
  File - Solution
*/

#ifndef __FILE_H__
#define __FILE_H__

template<class T>
class File {
 public:
  File();
  ~File();

  void enfiler(const T&);
  T defiler();
  void defiler(T&);
  const T& getTete() const;
  T& getTete();
  const T& getQueue() const;
  T& getQueue();

  void vider();

  bool vide() const { return queue == nullptr; }

  const File<T>& operator=(const File<T>&);

 private:
  File(const File<T>&);  // pas utilis

  class Cellule {
   public:
    Cellule(const T& e, Cellule* c = nullptr);
    T element;
    Cellule* suivante;
  };

  Cellule* queue;
};

/*
 Puisque File<T> est une classe gnrique, toutes ses dfinitions doivent tre
 inclues aprs les dclarations. On peut quand mme sparer les dclarations et
 les dfinitions en deux fichiers (file.h et file.hcc),  condition d'inclure le
 deuxime (file.hcc)  la fin du premier (file.h). Ainsi, un fichier source
 (.cc, .cpp, c++) qui inclut file.h inclura indirectement aussi file.hcc.

*/

// #include "file.hcc"
/**** Dbut des dfinitions pouvant tre mises dans file.hcc. ****/

#include <assert.h>

template<class T>
File<T>::File() {
  queue = nullptr;
}

template<class T>
File<T>::~File() {
  vider();
}

template<class T>
void File<T>::vider() {
  Cellule* fin = queue;
  while (queue != nullptr) {
    Cellule* suivante = queue->suivante;
    if (suivante == fin) suivante = nullptr;
    delete queue;
    queue = suivante;
  }
}

template<class T>
const T& File<T>::getTete() const {
  assert(queue != nullptr);
  return queue->suivante->element;
}

template<class T>
T& File<T>::getTete() {
  assert(queue != nullptr);
  return queue->suivante->element;
}

template<class T>
T& File<T>::getQueue() {
  assert(queue != nullptr);
  return queue->element;
}
template<class T>
const T& File<T>::getQueue() const {
  assert(queue != nullptr);
  return queue->element;
}

template<class T>
File<T>::Cellule::Cellule(const T& e, Cellule* c) : element(e), suivante(c) {}

template<class T>
void File<T>::enfiler(const T& e) {
  if (queue == nullptr) {
    queue = new Cellule(e);
    queue->suivante = queue;
  } else
    queue = queue->suivante = new Cellule(e, queue->suivante);
  assert(queue);
}

template<class T>
T File<T>::defiler() {
  assert(queue != nullptr);
  Cellule* c = queue->suivante;
  T e = c->element;

  if (queue == c)
    queue = nullptr;
  else
    queue->suivante = c->suivante;
  delete c;
  return e;
}

template<class T>
void File<T>::defiler(T& r) {
  assert(queue != nullptr);
  Cellule* c = queue->suivante;
  r = c->element;

  if (queue == c)
    queue = nullptr;
  else
    queue->suivante = c->suivante;

  delete c;
}

// Il existe une version unifie et plus courte de l'oprateur =.
// Cependant, la prsente version est plus  comprhensible .
template<class T>
const File<T>& File<T>::operator=(const File<T>& autre) {
  if (&autre == this) return *this;

  Cellule* c = queue ? queue->suivante : nullptr;
  Cellule* t = autre.queue ? autre.queue->suivante : nullptr;
  Cellule* d = nullptr;
  T temp;

  // 1. Rutiliser les cellules existantes de la File courante (*this)
  //    en crasant leur contenu par le contenu des cellules de la File
  //    autre.
  while (t) {
    temp = t->element;
    t = (t == autre.queue) ? nullptr : t->suivante;
    if (c != nullptr) {
      c->element = temp;
      d = c;
      c = (c == queue) ? nullptr : c->suivante;
    } else {
      // 2. Si la File courante (*this) davantage d'lments, alors
      //    enfiler les lments restant de la File autre.
      enfiler(temp);
    }
  }

  // 3. Si la File autre a moins d'lments, dtruire les cellules en trop.
  if (c != nullptr) {
    if (d != nullptr) d->suivante = queue->suivante;
    while (c) {
      t = c;
      c = queue == c ? nullptr : c->suivante;
      delete t;
    }
    queue = d;
  }
  return *this;
}

/**** Fin des dfinitions pouvant tre mises dans file.hcc. ****/

#endif
